///////////////////////////////////////////////////////////////////////////////
//
//    (C) Copyright 1995 - 2007 OSR Open Systems Resources, Inc.
//    All Rights Reserved
//
//    This sofware is supplied for instructional purposes only.
//
//    OSR Open Systems Resources, Inc. (OSR) expressly disclaims any warranty
//    for this software.  THIS SOFTWARE IS PROVIDED  "AS IS" WITHOUT WARRANTY
//    OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING, WITHOUT LIMITATION,
//    THE IMPLIED WARRANTIES OF MECHANTABILITY OR FITNESS FOR A PARTICULAR
//    PURPOSE.  THE ENTIRE RISK ARISING FROM THE USE OF THIS SOFTWARE REMAINS
//    WITH YOU.  OSR's entire liability and your exclusive remedy shall not
//    exceed the price paid for this material.  In no event shall OSR or its
//    suppliers be liable for any damages whatsoever (including, without
//    limitation, damages for loss of business profit, business interruption,
//    loss of business information, or any other pecuniary loss) arising out
//    of the use or inability to use this software, even if OSR has been
//    advised of the possibility of such damages.  Because some states/
//    jurisdictions do not allow the exclusion or limitation of liability for
//    consequential or incidental damages, the above limitation may not apply
//    to you.
//
//    OSR Open Systems Resources, Inc.
//    105 Route 101A Suite 19
//    Amherst, NH 03031  (603) 595-6500 FAX: (603) 595-6503
//    email bugs to: bugs@osr.com
//
//
//    MODULE:
//
//        OSRMEMDRV.C
//
//    ABSTRACT:
//
//      This file contains the skeleton code for a device driver
//      that demonstrates the techniques described in the NTINSIDER article
//      Sharing Memory Between Drivers and Applications , Revisited.
//
//    AUTHOR(S):
//
//        OSR Open Systems Resources, Inc.
// 
//    REVISION:   
//
//
///////////////////////////////////////////////////////////////////////////////

#include <ntddk.h>
#include "osrmemdrv.h"

#define OSRMEMDRV_LEVEL_VERBOSE 1

//
// Forward Declarations
//
// The following function appear in this file:
//
extern "C" NTSTATUS
DriverEntry(PDRIVER_OBJECT DriverObj, PUNICODE_STRING RegistryPath);

NTSTATUS OsrMemDrvDeviceControl(PDEVICE_OBJECT DeviceObject, PIRP Irp);
NTSTATUS OsrMemDrvCreate(PDEVICE_OBJECT DeviceObject, PIRP Irp);
NTSTATUS OsrMemDrvClose(PDEVICE_OBJECT DeviceObject, PIRP Irp);
NTSTATUS OsrMemDrvCleanup(PDEVICE_OBJECT DeviceObject, PIRP Irp);
NTSTATUS OsrMemDrvWrite(PDEVICE_OBJECT DeviceObject, PIRP Irp);
NTSTATUS OsrMemDrvRead(PDEVICE_OBJECT DeviceObject, PIRP Irp);
VOID     OsrMemDrvUnload(PDRIVER_OBJECT DriverObject);
NTSTATUS CreateAndMapMemory(PMDL* PMemMdl,PVOID* UserVa);
void     UnMapAndFreeMemory(PMDL PMdl,PVOID UserVa);
VOID     CancelSetShareMemRoutine(IN PDEVICE_OBJECT DeviceObject,IN PIRP Irp);

///////////////////////////////////////////////////////////////////////////////
//
//  DriverEntry
//
//    This routine is called by Windows when the driver is first loaded.  It is the
//    responsibility of this routine to find it's device and create whatever
//    device objects it needs.
//
//  INPUTS:
//
//      DriverObj - Address of the DRIVER_OBJECT created by Windows for this driver.
//
//      RegistryPath - UNICODE_STRING which represents this drivers KEY in the
//                   Registry.  
//
//  OUTPUTS:
//
//      None.
//
//  RETURNS:
//
//      STATUS_SUCCESS. Otherwise an error indicating why the driver could not
//                    Load.
//
//  IRQL:
//
//    This routine is called at IRQL_PASSIVE_LEVEL.
//
//  NOTES:
//
//
///////////////////////////////////////////////////////////////////////////////
NTSTATUS DriverEntry(PDRIVER_OBJECT DriverObj, PUNICODE_STRING RegistryPath)
{
    NTSTATUS                code;
    POSRMEMDRV_DEVICE_EXT   devExt;
    PDEVICE_OBJECT          devObj;
    UNICODE_STRING          devName, linkName;

    DbgPrintEx(DPFLTR_IHVDRIVER_ID,OSRMEMDRV_LEVEL_VERBOSE,
            "\nOSR OSRMEMDRV Driver -- Compiled %s %s\n",__DATE__, __TIME__);

    //
    // Establish dispatch entry points for the functions we support
    //

    //
    // Set up our dispatch routines...., notice we use cleanup
    //
    DriverObj->MajorFunction[IRP_MJ_CREATE]         =  OsrMemDrvCreate;
    DriverObj->MajorFunction[IRP_MJ_CLOSE]          =  OsrMemDrvClose;
    DriverObj->MajorFunction[IRP_MJ_CLEANUP]        =  OsrMemDrvCleanup;

    DriverObj->MajorFunction[IRP_MJ_READ]           =  OsrMemDrvRead;
    DriverObj->MajorFunction[IRP_MJ_WRITE]          =  OsrMemDrvWrite;
    DriverObj->MajorFunction[IRP_MJ_DEVICE_CONTROL] =  OsrMemDrvDeviceControl;

    //
    // Unload function
    //
    DriverObj->DriverUnload = OsrMemDrvUnload;

    //
    // Initialize the UNICODE device name.  This will be the "native NT" name
    // for our device.
    //
    RtlInitUnicodeString(&devName, L"\\Device\\OsrMemDrv");

    //
    // Ask the I/O Manager to create the device object and
    // device extension
    //
    code = IoCreateDevice(DriverObj,
                    sizeof(OSRMEMDRV_DEVICE_EXT),
                    &devName,
                    FILE_DEVICE_OSRMEMDRV,
                    FILE_DEVICE_SECURE_OPEN,       
                    FALSE,
                    &devObj);


    if(!NT_SUCCESS(code))  {

        DbgPrintEx(DPFLTR_IHVDRIVER_ID,OSRMEMDRV_LEVEL_VERBOSE,
                    "IoCreateDevice failed.  Status = 0x%0x\n", code);
        return(STATUS_UNSUCCESSFUL);
    }    

    //
    // Get a pointer to our device extension
    //
    devExt = (POSRMEMDRV_DEVICE_EXT)devObj->DeviceExtension;

    //
    // Save the device object pointer away for future reference... Not
    // that we ever reference it in the OSRMEMDRV driver.
    //
    devExt->DeviceObject = devObj;

    //
    //  Initialize MemUserList information.
    //
    ExInitializeFastMutex(&devExt->MemUserListMutex);
    InitializeListHead(&devExt->MemUserList);

    //
    //  Initialize IrpQueue information.
    //
    ExInitializeFastMutex(&devExt->IrpQueueMutex);
    InitializeListHead(&devExt->IrpQueue);

    //
    // Next, make the device accessible from user-mode applications.
    // Note that this name can be either the same or different from
    // the native "kernel mode" name of the device object, given above.
    //
    RtlInitUnicodeString(&linkName, L"\\??\\OsrMemDrv");

    code = IoCreateSymbolicLink(&linkName, &devName);

    if (!NT_SUCCESS(code)) {

        DbgPrintEx(DPFLTR_IHVDRIVER_ID,OSRMEMDRV_LEVEL_VERBOSE,
                    "IoCreateSymbolicLink failed.  Status = 0x%x\n", code);
        IoDeleteDevice(devObj);

        //
        // Indicate load failure to the I/O manager; driver image is deleted...
        //
        return(code);
    }


    //
    // Tell the I/O Manger we want MDLS for our reads and writes
    //
    devObj->Flags |= DO_DIRECT_IO;

    DbgPrintEx(DPFLTR_IHVDRIVER_ID,OSRMEMDRV_LEVEL_VERBOSE,
                            "OsrMemDrvDriverEntry: Leaving\n");

    return(STATUS_SUCCESS);

}


///////////////////////////////////////////////////////////////////////////////
//
//  OsrMemDrvCreate
//
//    This is the dispatch entry point for processing IRP_MJ_CREATE
//    functions.  When we get called here we allocate a block of memory that
//    will be used by all users of a particular file object for this
//    process.    The user can get a pointer to this shared memory
//    when they issue a IOCTL_OSR_GETMEM call.
//
//
//  INPUTS:
//
//    DeviceObject - Address of the DEVICE_OBJECT for our device.
//  
//    Irp - Address of the IRP representing the IRP_MJ_CREATE
//          call.
//
//  OUTPUTS:
//
//      None.
//
//  RETURNS:
//
//    STATUS_SUCCESS.   We never fail this function.        
//
//  IRQL:
//
//    This routine is called at IRQL_PASSIVE_LEVEL.
//
//  NOTES:
//
///////////////////////////////////////////////////////////////////////////////
NTSTATUS OsrMemDrvCreate(PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
    PIO_STACK_LOCATION      ioStackLocation;
    POSRMEMDRV_DEVICE_EXT   devExt = (POSRMEMDRV_DEVICE_EXT) DeviceObject->DeviceExtension;
    PMEM_USER_ENTRY         pEntry = NULL;
    NTSTATUS                status = STATUS_SUCCESS;
    BOOLEAN                 bFound = FALSE;

    ioStackLocation = IoGetCurrentIrpStackLocation(Irp);

    DbgPrintEx(DPFLTR_IHVDRIVER_ID,OSRMEMDRV_LEVEL_VERBOSE,"OsrMemDrvCreate: Starting\n");

    //
    // Lock access to our list of memory that was keep track of memory allocated
    // per create call.
    //
    ExAcquireFastMutex(&devExt->MemUserListMutex);

    //
    // Allocate an entry from paged pool, since we only access this memory at 
    // IRQL < DISPATCH_LEVEL.
    //
    pEntry = (PMEM_USER_ENTRY) ExAllocatePoolWithTag(PagedPool,sizeof(MEM_USER_ENTRY),'eumP');

    //
    // Make sure we allocated memory successfully, if not return an error.
    //
    if(!pEntry) {
        status = STATUS_INSUFFICIENT_RESOURCES;
    } else {
        //
        // Make sure we are running in the thread of the caller, since we cannot map
        // memory into the user's process if we are not running in that process.
        //
        if(PsGetCurrentThread() == Irp->Tail.Overlay.Thread) {
            //
            // Create and map the memory that we will user with our user.
            //
            status = CreateAndMapMemory(&pEntry->PMdl,&pEntry->UserVaToReturn);
            if(NT_SUCCESS(status)) {
                //
                // Creation was successful, note it and put the entry on our list.
                //
                pEntry->FileObject = ioStackLocation->FileObject;
                InsertTailList(&devExt->MemUserList,&pEntry->ListEntry);
            } else {
                //
                // Opps, didn't work, deallocate the allocated memory.
                //
                ExFreePool(pEntry);
            }
        } else {
            //
            // Not in the right context, so exit....
            //
            ExFreePool(pEntry);
            status = STATUS_ACCESS_DENIED;
        }
    }

    //
    // Release access to the list.
    //
    ExReleaseFastMutex(&devExt->MemUserListMutex);

    //
    // OsrMemDrv much to do....
    //

    Irp->IoStatus.Status = status;
    Irp->IoStatus.Information = 0;

    IoCompleteRequest(Irp, IO_NO_INCREMENT);

    DbgPrintEx(DPFLTR_IHVDRIVER_ID,OSRMEMDRV_LEVEL_VERBOSE,"OsrMemDrvCreate: Leaving\n");

    return(STATUS_SUCCESS);
}

///////////////////////////////////////////////////////////////////////////////
//
//  OsrMemDrvClose
//
//    This is the dispatch entry point for processing IRP_MJ_CLOSE functions.  
//    Since this is a simple device driver,
//    there's really nothing to do here by complete the requests with success.
//
//
//  INPUTS:
//
//    DeviceObject - Address of the DEVICE_OBJECT for our device.
//  
//    Irp - Address of the IRP representing the IRP_MJ_CLOSE
//          call.
//
//  OUTPUTS:
//
//      None.
//
//  RETURNS:
//
//    STATUS_SUCCESS.   We never fail this function.        
//
//  IRQL:
//
//    This routine is called at IRQL_PASSIVE_LEVEL.
//
//  NOTES:
//
///////////////////////////////////////////////////////////////////////////////
NTSTATUS OsrMemDrvClose(PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
    DbgPrintEx(DPFLTR_IHVDRIVER_ID,OSRMEMDRV_LEVEL_VERBOSE,"OsrMemDrvClose: Starting\n");

    //
    // OsrMemDrv much to do....
    //
    Irp->IoStatus.Status = STATUS_SUCCESS;
    Irp->IoStatus.Information = 0;

    IoCompleteRequest(Irp, IO_NO_INCREMENT);

    DbgPrintEx(DPFLTR_IHVDRIVER_ID,OSRMEMDRV_LEVEL_VERBOSE,"OsrMemDrvClose: Leaving\n");

    return(STATUS_SUCCESS);
}

///////////////////////////////////////////////////////////////////////////////
//
//  OsrMemDrvCleanup
//
//    This is the dispatch entry point for processing IRP_MJ_CLEANUP functions.  
//    Here we must do cleanup on any shared memory that we are using with the
//    application that is cleaning up.
//
//
//  INPUTS:
//
//    DeviceObject - Address of the DEVICE_OBJECT for our device.
//  
//    Irp - Address of the IRP representing the IRP_MJ_CLEANUP
//          call.
//
//  OUTPUTS:
//
//      None.
//
//  RETURNS:
//
//    STATUS_SUCCESS.   We never fail this function.        
//
//  IRQL:
//
//    This routine is called at IRQL_PASSIVE_LEVEL.
//
//  NOTES:
//
///////////////////////////////////////////////////////////////////////////////
NTSTATUS OsrMemDrvCleanup(PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
    PIO_STACK_LOCATION      ioStackLocation;
    POSRMEMDRV_DEVICE_EXT   devExt = (POSRMEMDRV_DEVICE_EXT) DeviceObject->DeviceExtension;
    PMEM_USER_ENTRY         pEntry = NULL;
    NTSTATUS                status = STATUS_SUCCESS;

    DbgPrintEx(DPFLTR_IHVDRIVER_ID,OSRMEMDRV_LEVEL_VERBOSE,"OsrMemDrvCleanup: Starting\n");

    ioStackLocation = IoGetCurrentIrpStackLocation(Irp);

    //
    // Gain access to the MemUserList...
    //
    ExAcquireFastMutex(&devExt->MemUserListMutex);

    //
    // Check to see if there are any entries on the list...
    //
    if(!IsListEmpty(&devExt->MemUserList)) {

        //
        // Go thru the list and remove the entry whose file object matches the one contained in the
        // cleanup IRP.
        //
        PLIST_ENTRY pLentry;
        for(pLentry = devExt->MemUserList.Flink; pLentry != &devExt->MemUserList; pLentry = pLentry->Flink) {
            pEntry = (PMEM_USER_ENTRY) pLentry;
            if(pEntry->FileObject == ioStackLocation->FileObject) {
                //
                // We have found a match, remove it from the list, unmap, free the memory, and then
                // deallocate the ENTRY from the list.   We're good after this...
                //
                RemoveEntryList(&pEntry->ListEntry);
                UnMapAndFreeMemory(pEntry->PMdl,pEntry->UserVaToReturn);
                ExFreePool(pEntry);
                break;
            }
        }
    }

    //
    // Release access to the list.
    //
    ExReleaseFastMutex(&devExt->MemUserListMutex);

    //
    // Gain access to the IRP queue....
    //
    ExAcquireFastMutex(&devExt->IrpQueueMutex);

    //
    // See if there are any hanging IOCTL_OSR_SETSHAREMEM Irps.
    //
    if(!IsListEmpty(&devExt->IrpQueue)) {

        //
        // The list is not empty..  Go through the list looking for an entry with a File Object that
        // matches the one contained in the cleanup IRP.
        //
        PLIST_ENTRY pLentry;
        for(pLentry = devExt->IrpQueue.Flink; pLentry != &devExt->IrpQueue; pLentry = pLentry->Flink) {
            pEntry = (PMEM_USER_ENTRY) pLentry;
            if(pEntry->FileObject == ioStackLocation->FileObject) {
                //
                // We found a match, remove the entry from the list,
                // clear the cancel routine,
                // Set the status of the hung irp to cancel and complete
                // it.   We're done with this list...
                // 
                RemoveEntryList(&pEntry->ListEntry);
                IoSetCancelRoutine(pEntry->Irp,NULL);
                pEntry->Irp->IoStatus.Status = STATUS_SUCCESS;
                pEntry->Irp->IoStatus.Information = 0;
                IoCompleteRequest(pEntry->Irp,IO_NO_INCREMENT);
                ExFreePool(pEntry);
                break;
            }
        }
    }

    //
    // Release access to the Irp Queue.
    //
    ExReleaseFastMutex(&devExt->IrpQueueMutex);

    //
    // OsrMemDrv much to do....
    //
    Irp->IoStatus.Status = STATUS_SUCCESS;
    Irp->IoStatus.Information = 0;

    IoCompleteRequest(Irp, IO_NO_INCREMENT);

    DbgPrintEx(DPFLTR_IHVDRIVER_ID,OSRMEMDRV_LEVEL_VERBOSE,"OsrMemDrvCleanup: Leaving\n");

    return(STATUS_SUCCESS);
}


///////////////////////////////////////////////////////////////////////////////
//
//  OsrMemDrvWrite
//
//    This is the write dispatch entry point for the driver, called when the
//    I/O Manager has an IRP_MJ_WRITE request for the driver to process.
//
//  INPUTS:
//
//      DeviceObject - Address of the DEVICE_OBJECT for our device.
//  
//    Irp - Address of the IRP representing the IRP_MJ_WRITE call.
//
//  OUTPUTS:
//
//      None.
//
//  RETURNS:
//
//    STATUS_PENDING, since we are putting the IRP on our internal queue.
//
//  IRQL:
//
//    This routine is called at IRQL_PASSIVE_LEVEL.
//
//  NOTES:
//
//    Since we set the DO_DIRECT_IO bit in the Device Object, all buffers
//    passed to us reside in the system's non-paged pool, a pointer to this
//    buffer is in Irp->MdlAddress.
//
///////////////////////////////////////////////////////////////////////////////
NTSTATUS OsrMemDrvWrite(PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
    PIO_STACK_LOCATION      ioStack = IoGetCurrentIrpStackLocation(Irp);
    POSRMEMDRV_DEVICE_EXT   devExt = (POSRMEMDRV_DEVICE_EXT) DeviceObject->DeviceExtension;
    BOOLEAN                 bFound = FALSE;
    PMEM_USER_ENTRY         pEntry;
    NTSTATUS                status = STATUS_INVALID_PARAMETER;

    DbgPrintEx(DPFLTR_IHVDRIVER_ID,OSRMEMDRV_LEVEL_VERBOSE,"OsrMemDrvWrite: Starting\n");

    //
    // Make sure that we received a valid request.
    //
    if(!Irp->MdlAddress || ioStack->Parameters.Write.Length < sizeof(OSR_MEMORY_SIZE)) {
        status = STATUS_INVALID_PARAMETER;
    } else {
        //
        // First check the IRP queue for an entry with a matching file object.
        //
        ExAcquireFastMutex(&devExt->IrpQueueMutex);

        if(!IsListEmpty(&devExt->IrpQueue)) {
            PLIST_ENTRY pLentry;
            for(pLentry = devExt->IrpQueue.Flink; pLentry != &devExt->IrpQueue; pLentry = pLentry->Flink) {
                pEntry = (PMEM_USER_ENTRY) pLentry;
                if(pEntry->FileObject == ioStack->FileObject) {
                    //
                    // Found an entry, get out of the loop.
                    //
                    bFound = TRUE;
                    break;
                }
            }
        } 

        //
        // See if an entry was found.
        //
        if(bFound) {
            //
            // We found an entry, see if we can get SVAs for the source and destination
            // buffer, and if so, copy the data.  otherwise return an error.
            //
            PUCHAR pDestChar = (PUCHAR) MmGetSystemAddressForMdlSafe(pEntry->Irp->MdlAddress,NormalPagePriority);
            PUCHAR pSrcChar = (PUCHAR) MmGetSystemAddressForMdlSafe(Irp->MdlAddress,NormalPagePriority);
            if(pDestChar && pSrcChar) {
                memcpy(pDestChar,pSrcChar,OSR_MEMORY_SIZE);
            } else {
                status = STATUS_INSUFFICIENT_RESOURCES;
            }
        }
        ExReleaseFastMutex(&devExt->IrpQueueMutex);

        //
        // See if Found is FALSE.  If it is, then we know we didn't find an IRP in the queue
        // above.
        //
        if(!bFound) {
            //
            // Search the MemUserList for an entry with a matching file object.
            //
            ExAcquireFastMutex(&devExt->MemUserListMutex);
            if(!IsListEmpty(&devExt->MemUserList)) {
                PLIST_ENTRY pLentry;
                for(pLentry = devExt->MemUserList.Flink; pLentry != &devExt->MemUserList; pLentry = pLentry->Flink) {
                    pEntry = (PMEM_USER_ENTRY) pLentry;
                    if(pEntry->FileObject == ioStack->FileObject) {
                        //
                        // We found a match exit.
                        //
                        bFound = TRUE;
                        break;
                    }
                }
            } 
            if(bFound) {
                //
                // We found a match if possible, copy the source to the destination...
                //
                PUCHAR pDestChar = (PUCHAR) MmGetSystemAddressForMdlSafe(pEntry->PMdl,NormalPagePriority);
                PUCHAR pSrcChar = (PUCHAR) MmGetSystemAddressForMdlSafe(Irp->MdlAddress,NormalPagePriority);
                if(pDestChar && pSrcChar) {
                    memcpy(pDestChar,pSrcChar,OSR_MEMORY_SIZE);
                } else {
                    status = STATUS_INSUFFICIENT_RESOURCES;
                }
            }
            ExReleaseFastMutex(&devExt->MemUserListMutex);
        }
    }

    //
    // Complete the IRP with success
    //
    Irp->IoStatus.Status = status;
    if(NT_SUCCESS(status)) {
        Irp->IoStatus.Information =  OSR_MEMORY_SIZE;
    }

    IoCompleteRequest(Irp, IO_NO_INCREMENT);

    DbgPrintEx(DPFLTR_IHVDRIVER_ID,OSRMEMDRV_LEVEL_VERBOSE,"OsrMemDrvWrite: Leaving\n");

    return(STATUS_SUCCESS);
}

///////////////////////////////////////////////////////////////////////////////
//
//  OsrMemDrvRead
//
//    This is the Read dispatch entry point for the driver, called when the
//    I/O Manager has an IRP_MJ_READ request for the driver to process.
//
//  INPUTS:
//
//    DeviceObject - Address of the DEVICE_OBJECT for our device.
//  
//    Irp - Address of the IRP representing the IRP_MJ_READ call.
//
//  OUTPUTS:
//
//      None.
//
//  RETURNS:
//
//    STATUS_PENDING, since we are putting the IRP on our internal queue.
//
//  IRQL:
//
//    This routine is called at IRQL_PASSIVE_LEVEL.
//
//  NOTES:
//
//    Since we set the DO_DIRECT_IO bit in the Device Object, all buffers
//    passed to us reside in the system's non-paged pool, a pointer to this
//    buffer is in Irp->MdlAddress.
//
///////////////////////////////////////////////////////////////////////////////
NTSTATUS OsrMemDrvRead(PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
    PIO_STACK_LOCATION      ioStack = IoGetCurrentIrpStackLocation(Irp);
    POSRMEMDRV_DEVICE_EXT   devExt = (POSRMEMDRV_DEVICE_EXT) DeviceObject->DeviceExtension;
    BOOLEAN                 bFound = FALSE;
    PMEM_USER_ENTRY         pEntry;
    NTSTATUS                status = STATUS_INVALID_PARAMETER;

    DbgPrintEx(DPFLTR_IHVDRIVER_ID,OSRMEMDRV_LEVEL_VERBOSE,"OsrMemDrvRead: Starting\n");

    //
    // Validate the input parameters.
    //
    if(!Irp->MdlAddress || ioStack->Parameters.Read.Length < sizeof(OSR_MEMORY_SIZE)) {
        status = STATUS_INVALID_PARAMETER;
    } else {
        //
        // See if an entry in the IRP Queue matches
        //
        ExAcquireFastMutex(&devExt->IrpQueueMutex);
        if(!IsListEmpty(&devExt->IrpQueue)) {
            PLIST_ENTRY pLentry;
            for(pLentry = devExt->IrpQueue.Flink; pLentry != &devExt->IrpQueue; pLentry = pLentry->Flink) {
                pEntry = (PMEM_USER_ENTRY) pLentry;
                if(pEntry->FileObject == ioStack->FileObject) {
                    //
                    // We found a hanging IRP, get out of the loop.
                    //
                    bFound = TRUE;
                    break;
                }
            }
        } 

        //
        // See if we found a match...
        //
        if(bFound) {
            //
            // We found a match, copy the source to the destination if possible.
            //
            PUCHAR pSrcChar = (PUCHAR) MmGetSystemAddressForMdlSafe(pEntry->Irp->MdlAddress,NormalPagePriority);
            PUCHAR pDestChar = (PUCHAR) MmGetSystemAddressForMdlSafe(Irp->MdlAddress,NormalPagePriority);
            if(pDestChar && pSrcChar) {
                memcpy(pDestChar,pSrcChar,OSR_MEMORY_SIZE);
            } else {
                status = STATUS_INSUFFICIENT_RESOURCES;
            }
        }
        ExReleaseFastMutex(&devExt->IrpQueueMutex);

        //
        // Check to see if an entry was found in the IRP queue.  If not, search the Memory List.
        //
        if(!bFound) {
            
            //
            // No hanging IRP found, so see if there is a match in the memory list.
            //
            ExAcquireFastMutex(&devExt->MemUserListMutex);
            if(!IsListEmpty(&devExt->MemUserList)) {
                PLIST_ENTRY pLentry;
                for(pLentry = devExt->MemUserList.Flink; pLentry != &devExt->MemUserList; pLentry = pLentry->Flink) {
                    pEntry = (PMEM_USER_ENTRY) pLentry;
                    if(pEntry->FileObject == ioStack->FileObject) {
                        //
                        // We found a match, exit.
                        //
                        bFound = TRUE;
                        break;
                    }
                }
            } 

            //
            // See if an entry was found.
            //
            if(bFound) {
                //
                // We found a match, copy the source to the destination if possible.
                //
                PUCHAR pSrcChar = (PUCHAR) MmGetSystemAddressForMdlSafe(pEntry->PMdl,NormalPagePriority);
                PUCHAR pDestChar = (PUCHAR) MmGetSystemAddressForMdlSafe(Irp->MdlAddress,NormalPagePriority);
                if(pDestChar && pSrcChar) {
                    memcpy(pDestChar,pSrcChar,OSR_MEMORY_SIZE);
                } else {
                    status = STATUS_INSUFFICIENT_RESOURCES;
                }
            }
            ExReleaseFastMutex(&devExt->MemUserListMutex);
        }
    }

    //
    // Complete the IRP 
    //
    Irp->IoStatus.Status = status;
    if(status == STATUS_SUCCESS) {
        Irp->IoStatus.Information =  OSR_MEMORY_SIZE;
    }

    IoCompleteRequest(Irp, IO_NO_INCREMENT);

    DbgPrintEx(DPFLTR_IHVDRIVER_ID,OSRMEMDRV_LEVEL_VERBOSE,"OsrMemDrvRead: Leaving\n");

    return(STATUS_SUCCESS);
}


///////////////////////////////////////////////////////////////////////////////
//
//  OsrMemDrvDeviceControl
//
//    This is the dispatch entry point for processing IRP_MJ_DEVICE_CONTROL
//    Irps.  
//
//
//  INPUTS:
//
//    DeviceObject - Address of the DEVICE_OBJECT for our device.
//  
//    Irp - Address of the IRP representing the IRP_MJ_DEVICE_CONTROL call.
//
//  OUTPUTS:
//
//      None.
//
//  RETURNS:
//
//    STATUS_SUCCESS if success, otherwise an appropriate status is returned.
//
//  IRQL:
//
//    This routine is called at IRQL_PASSIVE_LEVEL.
//
//  NOTES:
//
//    We presently only support IOCTL_OSR_OSRMEMDRV.
//
///////////////////////////////////////////////////////////////////////////////
NTSTATUS OsrMemDrvDeviceControl(PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
    ULONG                   operation;
    POSRMEMDRV_DEVICE_EXT   devExt = (POSRMEMDRV_DEVICE_EXT) DeviceObject->DeviceExtension;
    PIO_STACK_LOCATION      ioStack;
    NTSTATUS                status = STATUS_INVALID_PARAMETER;

    DbgPrintEx(DPFLTR_IHVDRIVER_ID,OSRMEMDRV_LEVEL_VERBOSE,"OsrMemDrvDeviceControl: Starting\n");

    //
    // Get the current IRP stack location
    //
    ioStack = IoGetCurrentIrpStackLocation(Irp);

    operation = ioStack->Parameters.DeviceIoControl.IoControlCode;

    //
    // Set the status field to zero for completion
    //
    Irp->IoStatus.Information = 0;

    switch (operation) {

        case IOCTL_OSR_SETSHAREMEM: {
            BOOLEAN bFound = FALSE;
            status = STATUS_SUCCESS;
            DbgPrintEx(DPFLTR_IHVDRIVER_ID,OSRMEMDRV_LEVEL_VERBOSE,
                            "OsrMemDrvDeviceControl: IOCTL_OSR_OSRMEMDRV processed\n");
            ExAcquireFastMutex(&devExt->IrpQueueMutex);
            if(!IsListEmpty(&devExt->IrpQueue)) {
                PLIST_ENTRY pLentry;
                PMEM_USER_ENTRY pEntry;
                for(pLentry = devExt->IrpQueue.Flink; pLentry != &devExt->IrpQueue; pLentry = pLentry->Flink) {
                    pEntry = (PMEM_USER_ENTRY) pLentry;
                    if(pEntry->FileObject == ioStack->FileObject) {
                        status = STATUS_ACCESS_DENIED;
                        bFound = TRUE;
                        break;
                    }
                }
            } 
            if(!bFound) {
                if(PsGetCurrentThread() == Irp->Tail.Overlay.Thread) {
                    PMEM_USER_ENTRY pEntry = (PMEM_USER_ENTRY) ExAllocatePoolWithTag(PagedPool,sizeof(MEM_USER_ENTRY),'eumI');
                    if(pEntry) {
                        pEntry->FileObject = ioStack->FileObject;
                        pEntry->Irp = Irp;
                        IoMarkIrpPending(Irp);
                        IoSetCancelRoutine(Irp,CancelSetShareMemRoutine);
                        InsertTailList(&devExt->IrpQueue,&pEntry->ListEntry);
                        ExReleaseFastMutex(&devExt->IrpQueueMutex);
                        return STATUS_PENDING;
                    } else {
                        status = STATUS_INSUFFICIENT_RESOURCES;
                    }  
                } else {
                    status = STATUS_ACCESS_DENIED;
                }
            }
            ExReleaseFastMutex(&devExt->IrpQueueMutex);
            }
            break;     

        case IOCTL_OSR_GETMEM: {
            PMEM_USER_ENTRY pEntry = NULL;
            DbgPrintEx(DPFLTR_IHVDRIVER_ID,OSRMEMDRV_LEVEL_VERBOSE,
                            "OsrMemDrvDeviceControl: IOCTL_OSR_GETMEM processed\n");

            ExAcquireFastMutex(&devExt->MemUserListMutex);

            if(!IsListEmpty(&devExt->MemUserList)) {
                PLIST_ENTRY pLentry;
                BOOLEAN     bFound = FALSE;
                status = STATUS_INVALID_PARAMETER;
                for(pLentry = devExt->MemUserList.Flink; pLentry != &devExt->MemUserList; pLentry = pLentry->Flink) {
                    pEntry = (PMEM_USER_ENTRY) pLentry;
                    if(pEntry->FileObject == ioStack->FileObject) {
                        //
                        // We have found a matching entry.   Return it to the caller if we can.
                        //
                        if(!Irp->AssociatedIrp.SystemBuffer || ioStack->Parameters.DeviceIoControl.OutputBufferLength <
                            sizeof(USER_MEMORY_ADDRESS)) {
                            status = STATUS_BUFFER_OVERFLOW;
                            Irp->IoStatus.Information = sizeof(USER_MEMORY_ADDRESS);
                        } else {
                            PUSER_MEMORY_ADDRESS pUMA = (PUSER_MEMORY_ADDRESS) Irp->AssociatedIrp.SystemBuffer;
                            pUMA->UserMemoryAddress = pEntry->UserVaToReturn;
                            status = STATUS_SUCCESS;
                            Irp->IoStatus.Information = sizeof(USER_MEMORY_ADDRESS);
                        }
                        break;
                    }
                }
            } else {
                status = STATUS_INVALID_PARAMETER;
            }
            ExReleaseFastMutex(&devExt->MemUserListMutex);
            }
            break;

        default:

            //
            // This is the case for a bogus request code.
            //
            status = STATUS_INVALID_PARAMETER;

            DbgPrintEx(DPFLTR_IHVDRIVER_ID,OSRMEMDRV_LEVEL_VERBOSE,
                "OsrMemDrvDeviceControl: Invalid IOCTL code 0x%0x\n", operation);

            break;
    }

    //
    // Complete the I/O Request
    //
    Irp->IoStatus.Status = status;

    IoCompleteRequest(Irp, IO_NO_INCREMENT);

    DbgPrintEx(DPFLTR_IHVDRIVER_ID,OSRMEMDRV_LEVEL_VERBOSE,"OsrMemDrvDeviceControl: Leaving\n");

    return(status);
}



///////////////////////////////////////////////////////////////////////////////
//
//  OsrMemDrvUnload
//
//    This routine is our dynamic unload entry point.  We are called here when
//    the OS wants to unload our driver.  It is our responsibility to release any
//    resources we allocated.
//
//  INPUTS:
//
//      DriverObj - Address of our DRIVER_OBJECT.
//
//  OUTPUTS:
//
//      None.
//
//  RETURNS:
//
//      None.
//
//  IRQL:
//
//    This routine is called at IRQL_PASSIVE_LEVEL.
//
//  NOTES:
//
///////////////////////////////////////////////////////////////////////////////
static VOID OsrMemDrvUnload(PDRIVER_OBJECT DriverObject)
{
    PDEVICE_OBJECT devObj;
    UNICODE_STRING linkName;
            
    DbgPrintEx(DPFLTR_IHVDRIVER_ID,OSRMEMDRV_LEVEL_VERBOSE,"OsrMemDrvUnload: Starting\n");

    //
    // For THIS driver, there will only ever be a single device object.
    // Because of this, we just get it from the DriverObj.  If this were
    // a multiple device driver, we would do this in a while loop...
    //
    devObj = DriverObject->DeviceObject;

    if (!devObj) {

        return;
    }

    RtlInitUnicodeString(&linkName, L"\\??\\OsrMemDrv");

    IoDeleteSymbolicLink(&linkName);

    //
    // Delete the device object
    //
    IoDeleteDevice(devObj);

    DbgPrintEx(DPFLTR_IHVDRIVER_ID,OSRMEMDRV_LEVEL_VERBOSE,"OsrMemDrvUnload: Leaving\n");
}

///////////////////////////////////////////////////////////////////////////////
//
//  CancelSetShareMemRoutine
//
//    This routine is our cancel routine for a queued IOCTL_OSR_SETSHAREMEM 
//    DeviceIoControl Irp.
//
//  INPUTS:
//
//      DeviceObject - Address of our DEVICE_OBJECT.
//      Irp - Address of the IRP to be canceled.
//
//  OUTPUTS:
//
//      None.
//
//  RETURNS:
//
//      None.
//
//  IRQL:
//
//    This routine is called at IRQL DISPATCH_LEVEL.
//
//  NOTES:
//
///////////////////////////////////////////////////////////////////////////////
VOID CancelSetShareMemRoutine(IN PDEVICE_OBJECT DeviceObject,IN PIRP Irp)
{
    POSRMEMDRV_DEVICE_EXT   devExt = (POSRMEMDRV_DEVICE_EXT) DeviceObject->DeviceExtension;
    PLIST_ENTRY             pLentry;
    PMEM_USER_ENTRY         pEntry;

    //
    // Release the Cancel SpinLock, we don't use it.
    //
    IoReleaseCancelSpinLock(Irp->CancelIrql);

    //
    // Gain access to the IrpQueue.
    //
    ExAcquireFastMutex(&devExt->IrpQueueMutex);

    //
    // Search the queue looking for the canceled IRP.
    //
    for(pLentry = devExt->IrpQueue.Flink; pLentry != &devExt->IrpQueue; pLentry = pLentry->Flink) {
        pEntry = (PMEM_USER_ENTRY) pLentry;
        if(pEntry->Irp == Irp) {
            //
            // We found the IRP.   Remove the entry from the list.
            // Do the cleanup, mark the irp as cancelled and complete it.
            // Delete the entry
            RemoveEntryList(&pEntry->ListEntry);
            IoSetCancelRoutine(pEntry->Irp,NULL);
            pEntry->Irp->IoStatus.Status = STATUS_CANCELLED;
            pEntry->Irp->IoStatus.Information = 0;
            IoCompleteRequest(pEntry->Irp,IO_NO_INCREMENT);
            ExFreePool(pEntry);
            break;
        }
    }

    //
    // Release access to the Queue
    //
    ExReleaseFastMutex(&devExt->IrpQueueMutex);
}


///////////////////////////////////////////////////////////////////////////////
//
//  CreateAndMapMemory
//
//    This routine allocates and maps the memory that we will share between a
//    process and our driver..
//
//  INPUTS:
//
//      None.
//
//  OUTPUTS:
//
//      PMemMdl - Address location to receive allocated MDL
//      UserVa - Address of location to receive UserVA of mapped memory
//
//  RETURNS:
//
//      STATUS_SUCCESS if it works, an error if it fails..
//
//  IRQL:
//
//    This routine is called at IRQL < DISPATCH_LEVEL.
//
//  NOTES:
//
//      We had better be running in the process where we want this 
//      memory mapped.
//
///////////////////////////////////////////////////////////////////////////////
NTSTATUS CreateAndMapMemory(PMDL* PMemMdl,PVOID* UserVa)
{
    PMDL                mdl;
    PVOID               userVAToReturn;
	PHYSICAL_ADDRESS    lowAddress;
	PHYSICAL_ADDRESS    highAddress;
	SIZE_T              totalBytes;

	//
	// Initialize the Physical addresses need for MmAllocatePagesForMdl
	//
	lowAddress.QuadPart = 0;
	highAddress.QuadPart = 0xFFFFFFFFFFFFFFFF;
	totalBytes = OSR_MEMORY_SIZE;
	
    //
    // Allocate a 4K buffer to share with the application
    //
	mdl = MmAllocatePagesForMdl(lowAddress,highAddress,lowAddress,totalBytes);
	
	if(!mdl) {
		return STATUS_INSUFFICIENT_RESOURCES;
    }

    //
    // The preferred V5 way to map the buffer into user space
    //
    userVAToReturn = 
         MmMapLockedPagesSpecifyCache(mdl,          // MDL
                                      UserMode,     // Mode
                                      MmCached,     // Caching
                                      NULL,         // Address
                                      FALSE,        // Bugcheck?
                                      NormalPagePriority); // Priority

    //
    // If we get NULL back, the request didn't work.
    // I'm thinkin' that's better than a bug check anyday.
    //
    if(!userVAToReturn)  {
        MmFreePagesFromMdl(mdl);        
        IoFreeMdl(mdl);
        return STATUS_INSUFFICIENT_RESOURCES;
    }

    //
    // Return the allocated pointers.
    //
    *UserVa = userVAToReturn;
    *PMemMdl = mdl;

    DbgPrintEx(DPFLTR_IHVDRIVER_ID,OSRMEMDRV_LEVEL_VERBOSE,"UserVA = 0x%0x\n", userVAToReturn);

    return STATUS_SUCCESS;
} 

///////////////////////////////////////////////////////////////////////////////
//
//  UnMapAndFreeMemory
//
//    This routine frees and unmaps the memory that we shared between a
//    process and our driver..
//
//  INPUTS:
//
//      PMemMdl - Address of allocated MDL
//      UserVa - Address of mapped UserVA 
//
//  OUTPUTS:
//
//      None.
//
//  RETURNS:
//
//      None.
//
//  IRQL:
//
//    This routine is called at IRQL < DISPATCH_LEVEL.
//
//  NOTES:
//
///////////////////////////////////////////////////////////////////////////////
void UnMapAndFreeMemory(PMDL PMdl,PVOID UserVa)
{
    //
    // Make sure we have an MDL to free.
    //
	if(!PMdl) {
		return;
	}
	
    //
    // Unmap the pages.
    //
	MmUnmapLockedPages(UserVa,PMdl);
	
    //
    // Free the pages from the MDL.
    //
	MmFreePagesFromMdl(PMdl);        

    //
    // Release the MDL.
    //
	IoFreeMdl(PMdl);
	
}
